from enum import IntEnum

# Defined in torch/csrc/cuda/shared/cudnn.cpp
is_cuda: bool

def getRuntimeVersion() -> tuple[int, int, int]: ...
def getCompileVersion() -> tuple[int, int, int]: ...
def getVersionInt() -> int: ...

class RNNMode(IntEnum):
    rnn_relu = ...
    rnn_tanh = ...
    lstm = ...
    gru = ...
